@@ -1,10 +1,12 @@
 ;;; smtpmail.el --- simple SMTP protocol (RFC 821) for sending mail
 
-;; Copyright (C) 1995, 1996 Free Software Foundation, Inc.
+;; Copyright (C) 1995, 1996, 2001 Free Software Foundation, Inc.
 
 ;; Author: Tomoji Kagatani <kagatani@rbc.ncl.omron.co.jp>
 ;; Maintainer: Brian D. Carlstrom <bdc@ai.mit.edu>
 ;; ESMTP support: Simon Leinen <simon@switch.ch>
+;; Hacked by Mike Taylor, 11th October 1999 to add support for
+;; automatically appending a domain to RCPT TO: addresses.
 ;; Keywords: mail
 
 ;; This file is part of GNU Emacs.
@@ -28,16 +30,14 @@
 
 ;; Send Mail to smtp host from smtpmail temp buffer.
 
-;; Please add these lines in your .emacs(_emacs).
+;; Please add these lines in your .emacs(_emacs) or use customize.
 ;;
-;;(setq send-mail-function 'smtpmail-send-it)
+;;(setq send-mail-function 'smtpmail-send-it) ; if you use `mail'
+;;(setq message-send-mail-function 'smtpmail-send-it) ; if you are using Gnus.
 ;;(setq smtpmail-default-smtp-server "YOUR SMTP HOST")
-;;(setq smtpmail-smtp-service "smtp")
 ;;(setq smtpmail-local-domain "YOUR DOMAIN NAME")
-;;(setq smtpmail-debug-info t)
-;;(load-library "smtpmail")
-;;(setq smtpmail-code-conv-from nil)
-;;(setq user-full-name "YOUR NAME HERE")
+;;(setq smtpmail-sendto-domain "YOUR DOMAIN NAME")
+;;(setq smtpmail-debug-info t) ; only to debug problems
 
 ;; To queue mail, set smtpmail-queue-mail to t and use 
 ;; smtpmail-send-queued-mail to send.
@@ -77,6 +77,22 @@ don't define this value."
   :type '(choice (const nil) string)
   :group 'smtpmail)
 
+(defcustom smtpmail-sendto-domain nil
+  "*Local domain name without a host name.
+This is appended (with an @-sign) to any specified recipients which do
+not include an @-sign, so that each RCPT TO address is fully qualified.
+\(Some configurations of sendmail require this.)
+
+Don't bother to set this unless you have get an error like:
+	Sending failed; SMTP protocol error
+when sending mail, and the *trace of SMTP session to <somewhere>*
+buffer includes an exchange like:
+	RCPT TO: <someone>
+	501 <someone>: recipient address must contain a domain
+"
+  :type '(choice (const nil) string)
+  :group 'smtpmail)
+
 (defcustom smtpmail-debug-info nil
   "*smtpmail debug info printout. messages and process buffer."
   :type 'boolean
@@ -99,10 +115,24 @@ and sent with `smtpmail-send-queued-mail'."
   :type 'directory
   :group 'smtpmail)
 
+(defcustom smtpmail-warn-about-unknown-extensions nil
+  "*If set, print warnings about unknown SMTP extensions.
+This is mainly useful for development purposes, to learn about
+new SMTP extensions that might be useful to support."
+  :type 'boolean
+  :version "21.1"
+  :group 'smtpmail)
+
 (defvar smtpmail-queue-index-file "index"
   "File name of queued mail index,
 This is relative to `smtpmail-queue-dir'.")
 
+(defvar smtpmail-address-buffer)
+(defvar smtpmail-recipient-address-list)
+
+;; Buffer-local variable.
+(defvar smtpmail-read-point)
+
 (defvar smtpmail-queue-index (concat smtpmail-queue-dir
 				     smtpmail-queue-index-file))
 
@@ -110,6 +140,7 @@ This is relative to `smtpmail-queue-dir'.")
 ;;;
 ;;;
 
+;;;###autoload
 (defun smtpmail-send-it ()
   (require 'mail-utils)
   (let ((errbuf (if mail-interactive
@@ -118,7 +149,11 @@ This is relative to `smtpmail-queue-dir'.")
 	(tembuf (generate-new-buffer " smtpmail temp"))
 	(case-fold-search nil)
 	delimline
-	(mailbuf (current-buffer)))
+	(mailbuf (current-buffer))
+	(smtpmail-code-conv-from
+	 (if enable-multibyte-characters
+	     (let ((sendmail-coding-system smtpmail-code-conv-from))
+	       (select-message-coding-system)))))
     (unwind-protect
 	(save-excursion
 	  (set-buffer tembuf)
@@ -230,7 +265,8 @@ This is relative to `smtpmail-queue-dir'.")
 			       smtpmail-queue-dir
 			       (concat (time-stamp-yyyy-mm-dd)
 				       "_" (time-stamp-hh:mm:ss))))
-		   (file-elisp (concat file-data ".el"))
+		      (file-data (convert-standard-filename file-data))
+		      (file-elisp (concat file-data ".el"))
 		   (buffer-data (create-file-buffer file-data))
 		   (buffer-elisp (create-file-buffer file-elisp))
 		   (buffer-scratch "*queue-mail*"))
@@ -274,7 +310,10 @@ This is relative to `smtpmail-queue-dir'.")
 						   (end-of-line)
 						   (point))))
 	(load file-msg)
-	(setq tembuf (find-file-noselect file-msg))
+	;; Insert the message literally: it is already encoded as per
+	;; the MIME headers, and code conversions might guess the
+	;; encoding wrongly.
+	(setq tembuf (find-file-noselect file-msg nil t))
 	(if (not (null smtpmail-recipient-address-list))
 	    (if (not (smtpmail-via-smtp smtpmail-recipient-address-list 
 					tembuf))
@@ -296,9 +335,16 @@ This is relative to `smtpmail-queue-dir'.")
       (concat (system-name) "." smtpmail-local-domain)
     (system-name)))
 
+(defun smtpmail-maybe-append-domain (recipient)
+  (if (or (not smtpmail-sendto-domain)
+	  (string-match "@" recipient))
+      recipient
+    (concat recipient "@" smtpmail-sendto-domain)))
+
 (defun smtpmail-via-smtp (recipient smtpmail-text-buffer)
   (let ((process nil)
-	(host smtpmail-smtp-server)
+	(host (or smtpmail-smtp-server
+		  (error "`smtpmail-smtp-server' not defined")))
 	(port smtpmail-smtp-service)
 	response-code
 	greeting
@@ -351,15 +397,15 @@ This is relative to `smtpmail-queue-dir'.")
 		      (throw 'done nil)))
 	      (let ((extension-lines (cdr (cdr response-code))))
 		(while extension-lines
-		  (let ((name (intern (downcase (substring (car extension-lines) 4)))))
+		  (let ((name (intern (downcase (car (split-string (substring (car extension-lines) 4) "[ ]"))))))
 		    (and name
 			 (cond ((memq name '(verb xvrb 8bitmime onex xone
 						  expn size dsn etrn
 						  help xusr))
 				(setq supported-extensions
 				      (cons name supported-extensions)))
-			       (t (message "unknown extension %s"
-					   name)))))
+			       (smtpmail-warn-about-unknown-extensions
+				(message "Unknown extension %s" name)))))
 		  (setq extension-lines (cdr extension-lines)))))
 
 	    (if (or (member 'onex supported-extensions)
@@ -438,7 +484,7 @@ This is relative to `smtpmail-queue-dir'.")
 	    ;; RCPT TO: <recipient>
 	    (let ((n 0))
 	      (while (not (null (nth n recipient)))
-		(smtpmail-send-command process (format "RCPT TO: <%s>" (nth n recipient)))
+		(smtpmail-send-command process (format "RCPT TO: <%s>" (smtpmail-maybe-append-domain (nth n recipient))))
 		(setq n (1+ n))
 
 		(setq response-code (smtpmail-read-response process))
@@ -559,8 +605,10 @@ This is relative to `smtpmail-queue-dir'.")
 (defun smtpmail-send-data-1 (process data)
   (goto-char (point-max))
 
-  (if (not (null smtpmail-code-conv-from))
-      (setq data (code-convert-string data smtpmail-code-conv-from *internal*)))
+  (if (and (multibyte-string-p data)
+	   smtpmail-code-conv-from)
+      (setq data (string-as-multibyte
+		  (encode-coding-string data smtpmail-code-conv-from))))
 	
   (if smtpmail-debug-info
       (insert data "\r\n"))
@@ -605,17 +653,16 @@ This is relative to `smtpmail-queue-dir'.")
 (defun smtpmail-deduce-address-list (smtpmail-text-buffer header-start header-end)
   "Get address list suitable for smtp RCPT TO: <address>."
   (require 'mail-utils)  ;; pick up mail-strip-quoted-names
-  (let
-      ((case-fold-search t)
-       (simple-address-list "")
-       this-line
-       this-line-end
-       addr-regexp)
     
-    (unwind-protect
-	(save-excursion
-	  ;;
-	  (set-buffer smtpmail-address-buffer) (erase-buffer)
+  (unwind-protect
+      (save-excursion
+	(set-buffer smtpmail-address-buffer) (erase-buffer)
+	(let
+	    ((case-fold-search t)
+	     (simple-address-list "")
+	     this-line
+	     this-line-end
+	     addr-regexp)
 	  (insert-buffer-substring smtpmail-text-buffer header-start header-end)
 	  (goto-char (point-min))
 	  ;; RESENT-* fields should stop processing of regular fields.
@@ -658,7 +705,7 @@ This is relative to `smtpmail-queue-dir'.")
 	    (setq smtpmail-recipient-address-list recipient-address-list))
 
 	  )
-      )
+	)
     )
   )
 
@@ -668,18 +715,15 @@ This is relative to `smtpmail-queue-dir'.")
 There may be multiple BCC: lines, and each may have arbitrarily
 many continuation lines."
   (let ((case-fold-search t))
-	(save-excursion (goto-char (point-min))
-	  ;; iterate over all BCC: lines
-	  (while (re-search-forward "^\(RESENT-\)?BCC:" header-end t)
-	        (delete-region (match-beginning 0) (progn (forward-line 1) (point)))
-		;; get rid of any continuation lines
-		(while (and (looking-at "^[ \t].*\n") (< (point) header-end))
-		  (replace-match ""))
-		)
-	  ) ;; save-excursion
-	) ;; let
-  )
-
+    (save-excursion
+      (goto-char (point-min))
+      ;; iterate over all BCC: lines
+      (while (re-search-forward "^\\(RESENT-\\)?BCC:" header-end t)
+	(delete-region (match-beginning 0)
+		       (progn (forward-line 1) (point)))
+	;; get rid of any continuation lines
+	(while (and (looking-at "^[ \t].*\n") (< (point) header-end))
+	  (replace-match ""))))))
 
 
 (provide 'smtpmail)
