@@ -34,7 +34,16 @@
 /* Cancel some #define's in usg5-4.h.
    Larry Hunter <hunter@nlm.nih.gov> said this was needed
    for Irix 6.5.  Let's see if it is safe in 6.N, N<5, as well.  */
-#undef TIOCSIGSEND
+/* Extrapolating from Irix 6.5, the problem is that (at least) the
+   bzero definition breaks what the X headers do.  The following means
+   that we lack prototypes for these functions, and we presumably lose
+   at least in the 64-bit ABI (though that's only supported on Irix
+   6.5, which I can test).  We may be saved by the fact that these
+   appear to be intrinsics in the SGI (Cray) compiler.  It's probably
+   appropriate to include strings.h here, but I can't test it.  See
+   irix6-5.h.  -- fx  */
 #undef bcopy
 #undef bcmp
 #undef bzero
+
+#undef TIOCSIGSEND
